# typing
from typing import Callable
from flask import Flask

METHODS = ['get', 'post', 'put', 'patch', 'options', 'delete']
BASE_URL = "/"

class QueryURLtoResultMapping:
    """ Mapping of a URL with query params to a result """

    def __init__(self, url_params:str, result:dict) -> None:
        self.url = f"{BASE_URL}?{url_params}"
        self.result = result

def call_wrapper(app:Flask, validator:Callable, url:str=BASE_URL, body:dict={}, methods:list=METHODS):
    """
        Wrapper to automate testing of multiple Flask methods for expected functionality.

        Arguments:
        - app {Flask} -- The Flask app under test
        - validator {Callable(str, flask.Response)} -- A function with assertions to run based on
            the passed config. Must accept two arguments, the method being tested and the response to 
            calling that method at the specified URL
        
        Keyword Arguments:
        - url {str} -- The URL to call with all passed methods. Defaults to "/"
        - body {dict} -- The request body to call with all passed methods. Defaults to {}
        - methods {List[str]} -- The methods to call the url with, defaults to :METHODS:

        Raises:
        - AssertionError: If an assertion in the passed validator fails
    """
    
    # Invoke a test client
    with app.test_client() as c:
        # For each specified method
        for method in methods:
            print(f"Calling [{method.upper()}] {url} with body: {body}: ", end='')
            # Call the method at the specified url with the passed body
            rv = getattr(c, method)(url, json=body)
            # Call the validator with the method and result to run assertions on the Request/Response
            validator(method, rv)
            
            print("ok!")



# typing
from .NormalizedRequest import NormalizedRequest

class RequestDataNormalizer:
    """ [Request Data Normalizer Flask Extension](https://github.com/Topazoo/Flask-Request-Data-Normalizer)
    
        To Use:
        ```
        app = Flask(__name__)
        RequestDataNormalizer(app)
        ```

        Overrides the functionality of a `flask.Request`:
        1. flask.Request.args - Base64 encoded query string parsing is supported via [QueryStringManager](https://github.com/Topazoo/Query-String-Manager)
        2. flask.Request.get_json - If the mimetype does not indicate JSON (:mimetype:`application/json`, see :attr:`is_json`), 
        this is will still attempt to parse JSON unless `force` is False
        3. flask.Request.get_json - Query parameters (`Request.args`), Form parameters (`Request.form`) and Body parameters are all returned.
        In vanilla Flask, this only returns Body parameters. The Body parameters take precedence if overlap occurs in keys.
    """

    def __init__(self, app=None):
        if app:
            self.init_app(app)

    def init_app(self, app):
        app.request_class = NormalizedRequest



# clients
from src.flask_request_data_normalizer import RequestDataNormalizer
from flask import Flask
import unittest

# utils
from tests.utils import QueryURLtoResultMapping, call_wrapper
from flask import request

# typing
from decimal import Decimal

class TestBase64QueryParams(unittest.TestCase):
    """
        Tests for Flask's ability to parse base64 encoded query strings in incoming
        requests to dictionaries:
            - New functionality for reading parsed base64 query strings from Request.args,
              Request.value, Request.get_json() and Request.json
    """

    app = Flask(__name__)
    my_extension = RequestDataNormalizer(app)
    app.testing = True

    def test_handle_request_with_basic_b64_query_string(self):
        """
        Test requests with one base64 encoded key/value pair in the query string
        """

        validate = QueryURLtoResultMapping("q=IkhlbGxvIg==", {"q": "Hello"})

        # Validations to run on each call
        def validator(_, __):
            assert request.args == validate.result
            assert request.values.to_dict() == validate.result
            assert request.json == validate.result
            assert request.get_json() == validate.result

        # Make calls to specified URL with specified methods with optional body
        call_wrapper(self.app, validator, validate.url)

    def test_handle_request_with_complex_b64_query_string(self):
        """
        Test requests with complex base64 encoded key/value pairs in the query string
        """

        validate = QueryURLtoResultMapping("data=eyJuZXN0ZWQiOnsiYSI6ImEiLCJiIjoiYiJ9LCJsaXN0IjpbMSx7ImluIjoibGlzdCJ9LHRydWVdfQ==", {
            'data': {
                'nested': {
                    'a': 'a', 'b': 'b'
                }, 
                'list': [1, {'in': 'list'}, True]
            }
        })

        # Validations to run on each call
        def validator(_, __):
            assert request.args == validate.result
            assert request.values.to_dict() == validate.result
            assert request.json == validate.result
            assert request.get_json() == validate.result

        # Make calls to specified URL with specified methods with optional body
        call_wrapper(self.app, validator, validate.url)

    def test_handle_request_with_mixed_b64_query_string(self):
        """
        Test requests with mixed base64 encoded and regular key/value pairs in the query string
        """

        validate = QueryURLtoResultMapping("hello%20world=eyJ0ZXN0IjogeyJuZXN0ZWQiOiAxfSwgInRlc3QyIjogW3siaGkiOiAiVGhlcmUifV0sICJ0ZXN0MyI6IDMuMTR9", {
            "hello world": {
                "test": {
                    "nested": 1
                }, 
                "test2": [{"hi": "There"}], "test3": Decimal("3.14")
            }
        })

        # Validations to run on each call
        def validator(_, __):
            assert request.args == validate.result
            assert request.values.to_dict() == validate.result
            assert request.json == validate.result
            assert request.get_json() == validate.result

        # Make calls to specified URL with specified methods with optional body
        call_wrapper(self.app, validator, validate.url)



# utils
from werkzeug.utils import cached_property
from QueryStringManager import QueryStringManager

# typing
from flask import Request
import typing as t

class NormalizedRequest(Request):
    """Works like the usual flask.Request with some added twists:
        1. flask.Request.args - Base64 encoded query string parsing is supported via [QueryStringManager](https://github.com/Topazoo/Query-String-Manager)
        2. flask.Request.get_json - If the mimetype does not indicate JSON (:mimetype:`application/json`, see :attr:`is_json`), 
        this is will still attempt to parse JSON unless `force` is False
        3. flask.Request.get_json - Query parameters (`Request.args`), Form parameters (`Request.form`) and Body parameters are all returned.
        In vanilla Flask, this only returns Body parameters. The Body parameters take precedence if overlap occurs in keys.
    """
    
    @cached_property
    def args(self) -> dict:
        """The parsed Query parameters (the part in the URL after the question
        mark) or None if no query parameters passed.

        Uses a [QueryStringManager](https://github.com/Topazoo/Query-String-Manager) to parse
        the query parameters including Base64 encoded query strings
        """

        if not self.query_string:
            return None

        return QueryStringManager.parse(self.query_string.decode())


    def get_json(self, force: bool = True, silent: bool = True, cache: bool = True) -> t.Optional[dict]:
        """Works like the usual get_json() with some added twists:
           1. Query parameters (`Request.args`), Form parameters (`Request.form`) and Body parameters are all present.
           In vanilla Flask, this only returns Body parameters. The Body parameters take precedence if overlap occurs in keys.
           2. Base64 encoded query string parsing is supported
           3. If the mimetype does not indicate JSON (:mimetype:`application/json`, see :attr:`is_json`), 
           this is will still attempt to parse JSON unless `force` is False

        If no data is present in the Query String, Form or Body, return None

        :param force: Ignore the mimetype and always try to parse JSON. (default: True)
        :param silent: Silence parsing errors and return None instead.
        :param cache: Store the parsed JSON to return for subsequent calls.
        """

        # Parse body data like usual
        body_data = super().get_json(force=force, silent=silent, cache=cache)

        # Parse query parameter data
        query_param_and_form_data = self.values.to_dict()

        # If no body data, no form data and no query params
        if not body_data and not query_param_and_form_data:
            return None
        
        return self.merge_request_data_dictionaries(query_param_and_form_data, body_data)
    

    def merge_request_data_dictionaries(self, dictionary1:dict, dictionary2:dict):
        """ Merge two dictionaries recursively. The value of the second
            dictionary will overwrite the first.
        """

        # Check "no-op" conditions
        if not dictionary1:
            return dictionary2
        if not dictionary2: 
            return dictionary1
        
        # For all keys and value in the second dictionary starting at top-level
        for key, value in dictionary2.items():
            # If the key is not on this level of the first dictionary, copy it over
            if key not in dictionary1:
                dictionary1[key] = value
            
            # Otherwise, if the key is in dictionary 1 on this level:
            else:
                # If the datatype of the value mismatches, default to the value of dict 2
                if type(value) != type(dictionary1[key]):
                    dictionary1[key] = value

                # If the datatype matches and is a dictionary, recurse
                elif isinstance(value, dict):
                    dictionary1[key] = self.merge_request_data_dictionaries(value, dictionary1[key])

                # If the datatype matches and is a list or something else, default to the value of dict 2
                else:
                    dictionary1[key] = value

        return dictionary1



# clients
from src.flask_request_data_normalizer import RequestDataNormalizer
from flask import Flask
import unittest

# utils
from tests.utils import QueryURLtoResultMapping, call_wrapper
from flask import request

# typing
from decimal import Decimal

class TestQueryParamsWithBody(unittest.TestCase):
    """
        Tests for Flask's ability to parse query strings and request bodies to dictionaries:
            - Combining both dictionaries with no overlaps
            - Combining both dictionaries with overlaps
    """

    app = Flask(__name__)
    my_extension = RequestDataNormalizer(app)
    app.testing = True

    def test_handle_request_with_query_param_and_body(self):
        """
        Test requests with one key/value in query params and the same in body params
        """

        validate = QueryURLtoResultMapping("test=val", {"test": "val"})
        body = {"body": "val"}

        # Validations to run on each call
        def validator(_, __):
            assert request.args == validate.result
            assert request.values.to_dict() == validate.result
            assert request.json == {**validate.result,  **body}
            assert request.get_json() ==  {**validate.result,  **body}

        # Make calls to specified URL with specified methods with optional body
        call_wrapper(self.app, validator, validate.url, body)
    
    def test_handle_request_with_complex_query_params_and_complex_body(self):
        """
        Test requests with mutiple keys/values in query params and body
        """

        validate = QueryURLtoResultMapping(
            "test=0.3&test2=True&test3=1&test4=-1&test5=-1a", 
            {
                "test": Decimal("0.3"),
                "test2": True,
                "test3": 1,
                "test4": -1,
                "test5": "-1a"
            })
        
        body = {
            "test6": "val",
            "sub": {
                "list": [
                    1,2,3, {
                        "deep_obj": None
                    }
                ]
            },
            "1": False
        }

        # Validations to run on each call
        def validator(_, __):
            assert request.args == validate.result
            assert request.values.to_dict() == validate.result
            assert request.json == {**body, **validate.result}
            assert request.get_json() == {**body, **validate.result}

        # Make calls to specified URL with specified methods with optional body
        call_wrapper(self.app, validator, validate.url, body)

    def test_handle_request_with_overlapping_query_params_body_params(self):
        """
        Test requests with mutiple keys/values in query params and body that overlap
        Body params take precedence
        """

        validate = QueryURLtoResultMapping(
            "test=0.3&test2=True&test3=1&test4=-1&test5=-1a", 
            {
                "test": Decimal("0.3"),
                "test2": True,
                "test3": 1,
                "test4": -1,
                "test5": "-1a"
            })
        
        body = {
            "test3": "val",
            "sub": {
                "list": [
                    1,2,3, {
                        "deep_obj": None
                    }
                ]
            },
            "test2": False
        }

        final = {
            "test": Decimal("0.3"),
            "test2": False,
            "test3": "val",
            "test4": -1,
            "test5": "-1a",
            "sub": {
                "list": [
                    1,2,3, {
                        "deep_obj": None
                    }
                ]
            }
        }

        # Validations to run on each call
        def validator(_, __):
            assert request.args == validate.result
            assert request.values.to_dict() == validate.result
            assert request.json == final
            assert request.get_json() == final

        # Make calls to specified URL with specified methods with optional body
        call_wrapper(self.app, validator, validate.url, body)